#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Dr. Yves-Laurent Kom Samo"
__copyright__ = "Copyright (C) 2020 KXY Technologies, Inc."
__license__ = """
Copyright (C) 2020 KXY TECHNOLOGIES, INC.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
__version__ = "0.3.3"

from kxy.api import *
from kxy.regression import *
from kxy.classification import *
from kxy.asset_management import *
from kxy.pandas_extension import *